/************************************************************************/
/*                                                                      */
/*                      Debouncing 8 Keys                               */
/*                      Sampling 4 Times                                */
/*                      With Repeat Function                            */
/*                                                                      */
/*              Author: Peter Dannegger                                 */
/*                      danni@specs.de                                  */
/*                                                                      */
/************************************************************************/
 
#include <stdint.h>
#include <avr/io.h>
#include <avr/interrupt.h>
 
#ifndef F_CPU
#define F_CPU           1000000                   // processor clock frequency
#endif
 
#define KEY_PORT        PORTD
#define KEY_PIN         PIND
#define KEY0            2
#define KEY1            3
#define KEY2            4
#define ALL_KEYS        (1<<KEY0 | 1<<KEY1 | 1<<KEY2)
 
#define LED_DDR         DDRB
#define LED_PORT        PORTB
#define LED0            0
#define LED1            1
#define LED2            2
 
uint8_t key_state;                                // debounced and inverted key state:
                                                  // bit = 1: key pressed
uint8_t key_press;                                // key press detect
 
ISR( TIMER0_OVF_vect )                            // every 10ms
{
  static uint8_t ct0, ct1;
  uint8_t i;
 
  TCNT0 = (uint8_t)(int16_t)-(F_CPU / 1024 * 10e-3 + 0.5);  // preload for 10ms
 
  i = key_state ^ ~KEY_PIN;                       // key changed ?
  ct0 = ~( ct0 & i );                             // reset or count ct0
  ct1 = ct0 ^ (ct1 & i);                          // reset or count ct1
  i &= ct0 & ct1;                                 // count until roll over ?
  key_state ^= i;                                 // then toggle debounced state
  key_press |= key_state & i;                     // 0->1: key press detect
 
}

 
///////////////////////////////////////////////////////////////////
//
// check if a key has been pressed. Each pressed key is reported
// only once
//
uint8_t get_key_press( uint8_t key_mask )
{
  cli();                                          // read and clear atomic !
  key_mask &= key_press;                          // read key(s)
  key_press ^= key_mask;                          // clear key(s)
  sei();
  return key_mask;
}
 

 
int main( void )
{

  uint8_t system_awake; 
  
  
  // Inputs
  KEY_PORT = 0xFF;  // enable pullups
  
  // Outputs
  LED_DDR  = 0xFF;   //set as output
  LED_PORT = 0xFF;
 
  TCCR0B = (1<<CS02)|(1<<CS00);			// divide by 1024
  TIMSK0 = 1<<TOIE0;				// enable timer interrupt
 
  system_awake = 0;
  sei();
 
  for(;;) {                                       // main loop
                                                  // single press
 
    if( ( get_key_press( 1<<KEY0 | 1<<KEY1) ) &&
        (0 == system_awake) 
      )
    {
      LED_PORT &= ~(1<<LED0);
      system_awake = 1;
    }
      
    if( ( get_key_press( 1<<KEY1 | 1<<KEY2 ) ) &&
        (1 == system_awake) 
      )
    {
      LED_PORT |=  (1<<LED0);
      system_awake = 0;
    }
 
  }
}
